import re
import os
import shutil
import getpass
import socket
import paramiko


class _SFTPClient(object):
    '''
    SFTP client object.

    Used to open an SFTP session across an open SSH .Transport and perform remote file operations.

    Instances of this class may be used as context managers.
    '''

    def __init__(self, ipaddr: str, port: int, user: str, passwd: str) -> None:
        self.ipaddr = ipaddr
        self.port = port
        self.user = user
        self.passwd = passwd

    def open(self):
        self.tran = paramiko.Transport((self.ipaddr, self.port))
        self.tran.connect(username=self.user, password=self.passwd)
        self.sftp = paramiko.SFTP.from_transport(self.tran)

    def close(self):
        self.sftp.close()
        self.tran.close()

    def put(self, localpath: str, remotepath: str):
        '''
        Copy a local file (localpath) to the SFTP server as remotepath.
        Any exception raised by operations will be passed through.
        This method is primarily provided as a convenience.

        The SFTP operations use pipelining for speed.

        Args:
            localpath: str, the local file to copy
            remotepath: str, the destination path on the SFTP server. Note
                        that the filename should be included.
                        Only specifying a directory may result in an error.

        Returns:
            an .SFTPAttributes object containing attributes about the given file
        '''
        assert isinstance(localpath, str)
        assert isinstance(remotepath, str)

        return self.sftp.put(localpath, remotepath)

    def get(self, remotepath: str, localpath: str):
        '''
        Copy a remote file (remotepath) from the SFTP server to the local host as localpath.
        Any exception raised by operations will be passed through.
        This method is primarily provided as a convenience.

        Args:
            remotepath: str, the remote file to copy
            localpath: str, the destination path on the local host
        '''
        assert isinstance(localpath, str)
        assert isinstance(remotepath, str)

        return self.sftp.get(remotepath, localpath)

    def exec(self, command):
        self.ssh = paramiko.SSHClient()
        self.ssh._transport = self.tran
        _, stdout, stderr = self.ssh.exec_command(command)
        data = self.__readline(stdout)
        if not data:
            data = self.__readline(stderr)
            if data:
                raise SCPException(data)
        return data

    def __readline(self, pipe) -> str:
        assert pipe

        lines = b""

        while True:
            try:
                data = pipe.read(1)
                if data:
                    lines += data
                else:
                    break
            except socket.timeout:
                pass

        return lines


class SCPCommandSt():
    def __init__(self, path=None, user=None, ipaddr=None):
        self.user = user
        self.ipaddr = ipaddr
        self.path = path


class SCPException(Exception):
    pass


class SCPCommandDef:
    DEFAULTPORT = 22
    OPTIONS = 'rP'
    # 递归 -r
    RECURSION = 1
    # 指定端口 -P
    PORT = 2
    # 1：本地拷贝文件到远端，0：远端拷贝文件到本地
    SENDTO = 4


class SCPCommand():
    def __init__(self) -> None:
        self.options = 0
        self.port = SCPCommandDef.DEFAULTPORT
        self.passwd = ''
        self.connection = None

    def __split(self, command: str) -> list:
        _subs = command.split()
        sub = ''
        subs = []
        for _sub in _subs:
            if _sub.endswith('\\') and not _sub.endswith('\\\\'):
                if sub:
                    sub = '{} {}'.format(sub[:-1], _sub)
                else:
                    sub = _sub
            else:
                if sub:
                    sub = '{} {}'.format(sub[:-1], _sub)
                else:
                    sub = _sub
                subs.append(sub)
                sub = ''
        return subs

    def __list_local_dir(self, path: str) -> list:
        files = []
        for file in os.listdir(path):
            fullpath = '{}{}{}'.format(path, os.sep, file)
            if os.path.isfile(fullpath):
                files.append(fullpath)
            elif os.path.isdir(fullpath):
                files.append(fullpath)
                files += self.__list_local_dir(fullpath)
        return files

    def __paser_path(self, path: str) -> SCPCommandSt:
        pattern = re.search(
            r'(((\w+){0,1})@((2[0-4]\d|25[0-5]|[1-9]?\d|1\d{2})\.){3}(2[0-4]\d|25[0-5]|[01]?\d\d?):)', path)
        if pattern:
            _user = pattern.group().split('@')[0]
            if not _user:
                _user = getpass.getuser()
            _ipaddr = pattern.group().lstrip('{}@'.format(_user)).rstrip(':')
            _path = path.lstrip('{}@{}:'.format(_user, _ipaddr))
            st_path = SCPCommandSt(_path, _user, _ipaddr)
        else:
            st_path = SCPCommandSt(path)
        return st_path

    def __scp(self, command: str) -> None:
        '''
        1. 远端 本地 本地 [本地] get
        2. 远端 远端 [本地] get
        3. 本地 本地 [本地] get
        4. 本地 本地 本地 [远端] put
        5. 本地 远端 [远端] errro
        6. 远端 [远端] error
        '''
        st_subs = []
        subs = self.__split(command)
        # 删除空元素
        if '' in subs:
            subs.remove('')

        if '-r' in subs:
            index = subs.index('-r')
            subs.pop(index)
            self.options |= SCPCommandDef.RECURSION

        # -P 参数
        if '-P' in subs:
            index = subs.index('-P')
            subs.pop(index)
            if not re.match(r'(^(6553[0-5]|655[0-2][0-9]|65[0-4]\d{2}|6[0-4]\d{3}|[1-5]\d{4}|[1-9]\d{0,3})$)',
                            subs[index]):
                raise SCPException('bad port "{}"'.format(subs[index]))
            self.port = int(subs.pop(index))

        # 解析参数，并判断合法性
        for sub in subs:
            if sub.startswith('-') and len(sub) > 1 and sub[1:] not in SCPCommandDef.OPTIONS:
                raise SCPException('unknown option -- {}\n'
                                   'usage: scp [-r] [-P port] source ... target [passwd]'.format(sub[1:]))

            st_sub = self.__paser_path(sub)
            if st_sub.ipaddr:
                # 如果最后一个为远端文件，则指令不完整，缺少密码
                if sub == subs[-1:][0]:
                    raise SCPException('{}: lost password'.format(command))
                # 如果存在远端文件，则最后一个为密码
                if not self.passwd:
                    self.passwd = subs.pop()
            st_subs.append(st_sub)

        if len(st_subs) < 2:
            raise SCPException(
                'usage: scp [-r] [-P port] source ... target [passwd]')

        # 取出dest，并从列表中删除
        st_dest = st_subs[-1]
        st_subs.remove(st_dest)

        # 判断本地的属性
        if not st_dest.ipaddr:
            if st_dest.path.endswith(os.sep):
                if os.path.isfile(st_dest.path):
                    raise SCPException(
                        "cp: failed to access '{}': Not a directory".format(st_dest.path))
                if not os.path.exists(st_dest.path):
                    raise SCPException(
                        "cp: cannot create regular file '{}': Not a directory".format(st_dest.path))
            else:
                if not os.path.exists(st_dest.path) and len(st_subs) > 1:
                    raise SCPException(
                        "{}: No such file or directory".format(st_dest.path))
                if not os.path.isdir(st_dest.path) and len(st_subs) > 1:
                    raise SCPException(
                        "{}: Not a directory".format(st_dest.path))
        else:
            self.options |= SCPCommandDef.SENDTO
            try:
                self.connection = _SFTPClient(
                    st_dest.ipaddr, self.port, st_dest.user, self.passwd)
                self.connection.open()
            except Exception:
                # 连接不成功
                raise SCPException(
                    "ssh: connect to host {} port {}: No route to host\n"
                    "lost connection".format(st_dest.ipaddr, self.port))

        # 解析子文件，并进行拷贝
        for st_sub in st_subs:
            # dest 是远端
            if self.options & SCPCommandDef.SENDTO:
                self.__put(st_sub, st_dest)

            # dest 是本地
            else:
                self.__get(st_sub, st_dest)

    def __put(self, localpath: SCPCommandSt, remotepath: SCPCommandSt):
        # 不支持源文件和目标文件都是远端文件，这里按照scp指令的错误来提示
        if localpath.ipaddr and self.options & SCPCommandDef.SENDTO:
            self.__close()
            raise SCPException(
                "Host key verification failed.\n"
                "lost connection")

        if not os.path.exists(localpath.path):
            self.__close()
            raise SCPException(
                '{}: No such file or directory'.format(localpath.path))

        if os.path.isdir(localpath.path) and (not self.options & SCPCommandDef.RECURSION):
            self.__close()
            raise SCPException(
                "cp: -r not specified; omitting directory '{}'".format(localpath.path))

        if os.path.isfile(localpath.path):
            filename = os.path.split(localpath.path)[1].lstrip('\\')
            self.connection.put(
                localpath.path, '{}/{}'.format(remotepath.path, filename))

        else:
            files = self.__list_local_dir(localpath.path)
            remotedir = localpath.path
            for file in files:
                if os.path.isdir(file):
                    relative = file.split(localpath.path)[1].replace('\\', '/')
                    remotedir = '{}/{}'.format(remotepath.path, relative)
                    self.connection.exec("mkdir -p {}".format(remotedir))
                else:
                    filename = os.path.split(file)[1].lstrip('\\')
                    remotefile = '{}/{}'.format(remotedir, filename)
                    self.connection.put(file, remotefile)

    def __get(self, remotepath: SCPCommandSt, localpath: SCPCommandSt):
        # 远端到本地
        if remotepath.ipaddr:
            if not self.connection:
                try:
                    self.connection = _SFTPClient(
                        remotepath.ipaddr, self.port, remotepath.user, self.passwd)
                    self.connection.open()
                except Exception:
                    # 连接不成功
                    raise SCPException(
                        "ssh: connect to host {} port {}: No route to host\n"
                        "lost connection".format(remotepath.ipaddr, self.port))
            isdir = 'file' in self.connection.exec('''function isdir()
                            {
                                if [ -d $1 ];then
                                    echo 'file'
                                fi
                            }
                            isdir %s
                            ''' % (remotepath.path)).decode('utf-8').split('\n')

            if isdir and (not self.options & SCPCommandDef.RECURSION):
                self.__close()
                raise SCPException(
                    "scp: -r not specified; omitting directory '{}'".format(remotepath.path))

            if isdir:
                subs = self.connection.exec('''function findfiles()
                            {
                                local root=$1
                                for file in `ls $root`
                                do
                                    if [ -d $root/$file ];then
                                        findfiles $root/$file
                                    else
                                        echo $root/$file
                                    fi
                                done
                            }
                            findfiles %s
                            ''' % (remotepath.path)).decode('utf-8').split('\n')
                for sub in subs:
                    if not sub:
                        continue

                    relative = sub.split(remotepath.path)[1]
                    localfile = '{}{}{}'.format(
                        localpath.path, os.sep, relative)
                    localdir = os.path.split(localfile)[0]
                    if not os.path.exists(localdir):
                        os.makedirs(localdir)
                    self.connection.get(sub, localfile)
            else:
                self.connection.get(remotepath.path, '{}{}{}'.format(
                    localpath.path, os.sep, os.path.split(remotepath.path)[1]))
        # 本地到本地
        else:
            if not os.path.exists(remotepath.path):
                self.__close()
                raise SCPException(
                    '{}: No such file or directory'.format(remotepath.path))

            if os.path.isdir(remotepath.path) and (not self.options & SCPCommandDef.RECURSION):
                self.__close()
                raise SCPException(
                    "cp: -r not specified; omitting directory '{}'".format(remotepath.path))

            if os.path.isfile(remotepath.path):
                shutil.copy2(remotepath.path, localpath.path)
            else:
                files = self.__list_local_dir(remotepath.path)
                localdir = localpath.path
                if not os.path.exists(localdir):
                    os.makedirs(localdir)
                for file in files:
                    print(file)
                    if os.path.isdir(file):
                        relative = file.split(remotepath.path)[1]
                        localdir = '{}{}{}'.format(
                            localpath.path, os.sep, relative)
                        if not os.path.exists(localdir):
                            os.makedirs(localdir)
                    else:
                        filename = os.path.split(file)[1]
                        shutil.copy2(file, '{}{}{}'.format(
                            localdir, os.sep, filename))

    def __close(self):
        if self.connection:
            self.connection.close()
            self.connection = None
        self.port = None
        self.passwd = None
        self.options = 0

    def __call__(self, command: str) -> bool:
        assert isinstance(command, str)
        self.__scp(command)
        self.__close()
