from __future__ import print_function
import os
import re
import ast
import json
import string
from ast import NodeVisitor, parse, iter_fields, get_docstring
from re import compile as regexpCompile, IGNORECASE
from os.path import getsize
from codecs import BOM_UTF8, open as codecsOpen
from chardet import detect
try:
    from runtime import Runtime
except Exception:
    class Runtime:
        @staticmethod
        def get_project_path():
            return '.'


class TransferDef:
    DOCS_STEP1 = 0
    DOCS_STEP2 = 1
    DOCS_STEP3 = 2
    DOCS_STEP4 = 3
    DOCS_STEP5 = 4

    MD_STEP1 = 0
    MD_STEP2 = 1
    MD_STEP3 = 2

    MD_SUB_STEP1 = 0
    MD_SUB_STEP2 = 1
    MD_SUB_STEP3 = 2
    MD_SUB_STEP4 = 3
    MD_SUB_STEP5 = 4

    LANGUAGE_CN = "CN"
    LANGUAGE_EN = "EN"
    LANGUAGE_UNKOWN = "UNKOWN"

    REMOVE_NOTA = u'[’·°–!"#$%&\'()*+,-./:;<=>?@，。?★、…【】（）《》？“”‘’！[\\]^_`{|}~]+'
    REMOVE_PUNCTUATION_MAP = dict((ord(char), None)
                                  for char in string.punctuation)

    MARKDOWN_CN_SKIP = ["描述", "厂家"]
    MARKDOWN_EN_SKIP = ["Description", "Vendor"]

    EN_JSON_FILE = "docstring_en.json"
    CN_JSON_FILE = "docstring_zh.json"


class BaseDocsTranser(NodeVisitor):
    def __init__(self, fullpath: str, module: str = None) -> None:
        self.fullpath = fullpath
        self.encoding = 'ascii'
        self.vaild_cn_docs = {}
        self.vaild_en_docs = {}
        self.all_cn_docs = []
        self.all_en_docs = []
        self.python_files = []
        self.md_files = []
        self.module = module

    def _get_files(self, path: str) -> list:
        files = os.listdir(path)

        # 获取模块名称
        if not self.module:
            if self.fullpath.endswith('/'):
                rstrip_code = '/'
            elif self.fullpath.endswith('\\'):
                rstrip_code = '\\'
            else:
                rstrip_code = ''
            self.module = os.path.split(self.fullpath.rstrip(rstrip_code))[1]

        # 查找文件
        for file in files:
            if file == '__init__.py':
                continue
            fullname = '{0}{1}{2}'.format(path, os.sep, file)
            if os.path.isdir(fullname):
                self._get_files(fullname)
            if fullname.endswith('.py'):
                self.python_files.append(fullname)
            if fullname.endswith('.md'):
                self.md_files.append(fullname)

        # 有makrdown的python，以markdwon为主
        for md_file in self.md_files:
            if "{}.py".format(md_file.rsplit('_cn.md')[0]) in self.python_files:
                self.python_files.remove(
                    "{}.py".format(md_file.rsplit('_cn.md')[0]))
            elif "{}.py".format(md_file.rsplit('_en.md')[0]) in self.python_files:
                self.python_files.remove(
                    "{}.py".format(md_file.rsplit('_en.md')[0]))
            elif "{}.py".format(md_file.rsplit('.md')[0]) in self.python_files:
                self.python_files.remove(
                    "{}.py".format(md_file.rsplit('.md')[0]))

    def _walk_ClassDef(self, node):
        functions_desc = {}
        ret_docs = {}
        assign = []
        bases = []
        ret_docs[node.name] = {}
        for base in node.bases:
            if hasattr(base, 'id'):
                bases.append(base.id)
        for _, value in iter_fields(node):
            if not isinstance(value, list):
                continue

            for n in value:
                if not isinstance(n, (ast.AsyncFunctionDef, ast.FunctionDef, ast.Assign)):
                    continue

                if isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef)):
                    func_doc = self._walk_FunctionDef(n)
                    if func_doc:
                        functions_desc.update(func_doc)
                else:
                    assign.append(self._walk_Assign(n))

        return {node.name: {'bases': bases, 'doc': get_docstring(node),
                            'functions': functions_desc, 'assign': assign, 'type': 'class'}}

    def _get_ast_dict_value(self, node):
        node_len = len(node.keys)
        value = {}
        for i in range(node_len):
            value[self._get_ast_value(node.keys[i])] = self._get_ast_value(
                node.values[i])
        return value

    def _get_ast_list_value(self, node):
        value = []
        for sub in node.elts:
            value.append(self._get_ast_value(sub))
        return value

    def _get_ast_tupe_value(self, node):
        return tuple(self._get_ast_list_value(node))

    def _get_ast_value(self, node):
        if isinstance(node, ast.Dict):
            return self._get_ast_dict_value(node)
        if isinstance(node, ast.List):
            return self._get_ast_list_value(node)
        if isinstance(node, ast.Tuple):
            return self._get_ast_tupe_value(node)
        if hasattr(node, 'value'):
            return node.value
        if hasattr(node, 'n'):
            return node.n
        if hasattr(node, 's'):
            return node.s
        if hasattr(node, 'id'):
            return node.id

    def _walk_Assign_values(self, node):
        if isinstance(node, ast.List):
            ret = []
            for value in node.elts:
                if isinstance(node, (ast.List, ast.Dict)):
                    ret.append(self._walk_Assign_values(value))
                else:
                    ret.append(self._get_ast_value(value))
            return ret
        elif isinstance(node, ast.Dict):
            ret = {}
            for i in range(len(node.keys)):
                ast_key = self._get_ast_value(node.keys[i])
                ast_value = self._get_ast_value(node.values[i])
                if ast_value:
                    ret[ast_key] = ast_value
                else:
                    ret[ast_key] = self._walk_Assign_values(node.values[i])
            return ret
        return self._get_ast_value(node)

    def _walk_Assign(self, node):
        '''
        Assign 是列表，子元素是列表和字典等
        '''
        assign_name = node.targets[0].id
        assign_value = self._walk_Assign_values(node.value)
        return {assign_name: assign_value}

    def _walk_arguments(self, node):
        args_len = len(node.args)
        defaults_len = len(node.defaults)
        defaults = []
        args_dict = {}
        # 补全默认参数
        if defaults_len < args_len:
            for i in range(args_len - defaults_len):
                defaults.append(None)
        for i in range(defaults_len):
            defaults.append(node.defaults[i])

        for i in range(args_len):
            if node.args[i].arg == 'self':
                continue

            arg_dict = {}
            # 参数类型解析
            if hasattr(node.args[i].annotation, 'id'):
                arg_dict['type'] = node.args[i].annotation.id
            elif hasattr(node.args[i].annotation, 'func') and hasattr(node.args[i].annotation.func, 'id'):
                arg_dict['type'] = node.args[i].annotation.func.id

            # 默认参数解析
            default = self._get_ast_value(defaults[i])
            if default:
                arg_dict['default'] = default

            args_dict[node.args[i].arg] = arg_dict

        return args_dict

    def _walk_FunctionDef(self, node):
        if node.name.startswith('_') and node.name != '__init__':
            return

        function_desc = {'doc': get_docstring(node)}
        function_desc['type'] = 'function'
        for _, value in iter_fields(node):
            if not isinstance(value, ast.arguments):
                continue

            if isinstance(value, ast.arguments):
                function_desc['args'] = self._walk_arguments(value)
        self._get_ast_value(node.returns)
        if hasattr(node.returns, 'id'):
            function_desc['returns'] = node.returns.id
        elif hasattr(node.returns, 'func') and hasattr(node.returns.func, 'id'):
            function_desc['returns'] = node.returns.func.id

        return {node.name: function_desc}

    def _walk_AsyncFunctionDef(self, node):
        return self._walk_FunctionDef(node)

    def _walk_Expr(self, node):
        pass

    def _walk_Module(self, node):
        module_dict = {}
        if get_docstring(node):
            # 模块注释说明
            pass

        for _, value in iter_fields(node):
            for n in value:
                if not isinstance(n, (ast.AsyncFunctionDef,
                                      ast.FunctionDef,
                                      ast.ClassDef,
                                      ast.Module,
                                      ast.Expr)):
                    continue

                result = getattr(self, '_walk_{}'.format(
                    n.__class__.__name__))(n)
                if result:
                    module_dict.update(result)

        return module_dict

    def _read_file(self, filename: str) -> list:
        numOfSampleBytes = getsize(filename)
        sampleBytes = open(filename, 'rb').read(numOfSampleBytes)
        sampleByteAnalysis = detect(sampleBytes)
        self.encoding = sampleByteAnalysis['encoding'] or 'ascii'

        # Switch to generic versions to strip the BOM automatically.
        if sampleBytes.startswith(BOM_UTF8):
            self.encoding = 'UTF-8-SIG'
        elif self.encoding.startswith("UTF-16"):
            self.encoding = "UTF-16"
        elif self.encoding.startswith("UTF-32"):
            self.encoding = "UTF-32"

        # Read contents of input file.
        if self.encoding == 'ascii':
            inFile = open(filename)
        else:
            inFile = codecsOpen(filename, encoding=self.encoding)

        lines = inFile.readlines()
        inFile.close()

        return ''.join(lines)

    def _walk_python(self, filename: str) -> None:
        data = self._read_file(filename)
        node = parse(data, filename)
        return self._walk_Module(node)

    def _get_function_desc(self, data: str):
        desc = []
        args = []
        ret = []
        example = []
        step = 'desc'
        for line in data.splitlines():
            line = line.strip()
            if not line:
                continue

            match = Docs2Markdown.__args_start_re.match(line)
            if match:
                if step != 'args':
                    step = 'args'
                    continue
            '''
            match = Docs2Markdown.__args_re.match(line)
            if match:
                args.append(line)
                if step != 'args':
                    step = 'args'
                continue
            '''

            match = Docs2Markdown.__returns_start_re.match(line)
            if match:
                if step != 'ret':
                    step = 'ret'
                    continue

            match = Docs2Markdown.__raises_start_re.match(line)
            if match:
                if step != 'raise':
                    step = 'raise'
                    continue

            match = Docs2Markdown.__examples_start_re.match(line)
            if match:
                if step != 'example':
                    step = 'example'
                    continue
            if step == 'desc':
                desc.append(line)
            elif step == 'args':
                args.append('{}'.format(line))
            elif step == 'ret':
                ret.append(line)
            elif step == 'example':
                example.append(line)

        return desc, args, ret, example

    def _sub_bases(self, docs, file_class, bases):
        _bases = []
        for base in bases:
            if base not in file_class:
                continue
            if 'bases' not in docs[file_class[base]][base]:
                continue
            if docs[file_class[base]][base]['bases']:
                _bases += self._sub_bases(docs, file_class,
                                          docs[file_class[base]][base]['bases'])
            if base not in _bases:
                _bases.insert(0, base)

        return _bases

    def _merge_sub_doc(self, docs: dict, class_file: dict) -> dict:
        assert isinstance(docs, dict)
        assert isinstance(class_file, dict)

        valid_docs = {}
        for file in docs:
            if os.path.split(file)[1].startswith('_'):
                continue
            valid_docs[file] = {}
            for _class in docs[file].keys():
                valid_docs[file][_class] = {}
                if 'functions' not in docs[file][_class]:
                    continue
                if 'bases' not in docs[file][_class]:
                    continue
                if not docs[file][_class]['bases']:
                    continue

                for base in docs[file][_class]['bases']:
                    if base not in class_file:
                        continue
                    if 'functions' not in docs[class_file[base]][base]:
                        continue
                    for function in docs[class_file[base]][base]['functions']:
                        if function not in docs[file][_class]['functions']:
                            docs[file][_class]['functions'][function] = docs[class_file[base]
                                                                             ][base]['functions'][function]
        return docs

    def _class_file(self, docs: dict) -> list:
        assert isinstance(docs, dict)

        class_file = {}
        for file in docs:
            for item in docs[file]:
                if 'functions' in docs[file][item]:
                    class_file[item] = file

        return class_file

    def _find_bases(self, docs: dict, class_file: dict) -> None:
        assert isinstance(docs, dict)
        assert isinstance(class_file, dict)

        for _class in class_file:
            if 'bases' in docs[class_file[_class]][_class] and docs[class_file[_class]][_class]['bases']:
                bases = self._sub_bases(
                    docs, class_file, docs[class_file[_class]][_class]['bases'])
                if not bases:
                    continue
                if _class.startswith('_') or _class.endswith('_'):
                    continue
                if os.path.split(class_file[_class])[1].startswith('_'):
                    continue
                docs[class_file[_class]][_class]['bases'] = bases

    def _check_md_files(self, files):
        _files = []
        for file in files:
            _path = os.path.split(file)[0]
            _file = os.path.split(file)[1]
            # if _file.startswith('_'):
            #    continue
            _basefile = '{}{}{}'.format(_path, os.sep, _file.split('.py')[0])
            if os.path.isfile('{}_cn.md'.format(_basefile)) or\
               os.path.isfile('{}_en.md'.format(_basefile)):
                continue
            _files.append(file)

        return _files


class Docs2Markdown(BaseDocsTranser):
    __args_start_re = regexpCompile(r"^(\s*(?:(?:Keyword\s+)?"
                                    r"(?:A|Kwa)rg(?:ument)?|Attribute)s?"
                                    r"\s*:\s*)$", IGNORECASE)
    # __args_re = regexpCompile(r"^\s*(?P<name>\w+)\s*(?P<type>\(?\S*\)?)?\s*"
    #                          r"(?:-|:)+\s+(?P<desc>.+)$")
    __returns_start_re = regexpCompile(
        r"^\s*(?:Return|Yield)s?:\s*$", IGNORECASE)
    __raises_start_re = regexpCompile(r"^\s*(Raises|Exceptions|See Also):\s*$",
                                      IGNORECASE)
    __examples_start_re = regexpCompile(r"^\s*(?:Example|Doctest)s?:\s*$",
                                        IGNORECASE)

    def __init__(self, fullpath: str, module: str = None) -> None:
        super().__init__(fullpath, module)

    def _write_data(self, fileno, data):
        fileno.write('{}\n'.format(data))

    def _write_module(self, fileno, data):
        self._write_data(fileno, '# Model')
        compatible = []
        for key in data:
            if 'assign' in data[key]:
                for assign in data[key]['assign']:
                    if isinstance(assign, dict):
                        if 'compatible' in assign:
                            compatible += assign['compatible']

        if compatible:
            for module in compatible:
                self._write_data(fileno, '* {}'.format(module))
        else:
            self._write_data(fileno, '* None')
        self._write_data(fileno, '')
        self._write_data(fileno, '')

    def _write_module_desc(self, fileno, data):
        self._write_data(fileno, '## Description')
        if data['module_md_desc']:
            self._write_data(fileno, data['module_md_desc'])
        else:
            self._write_data(fileno, '* None')
        self._write_data(fileno, '')

    def _write_vendor(self, fileno, data):
        self._write_data(fileno, '## Vendor')
        vendor = ''
        for key in data:
            if 'attachment' in data[key]:
                for attachment in data[key]['attachment']:
                    if isinstance(attachment, dict):
                        if 'vendor' in attachment:
                            vendor = attachment['vendor']

        if vendor:
            self._write_data(fileno, '{}'.format(vendor))
        else:
            self._write_data(fileno, '* None')
        self._write_data(fileno, '')

    def _write_class(self, fileno, data):
        self._write_data(fileno, '## Class Name')
        classes = []
        for key in data:
            if 'functions' in data[key]:
                classes.append(key)

        if classes:
            for _class in classes:
                self._write_data(fileno, '* {}'.format(_class))
                # Class method
                self._write_class_function(fileno, data[_class]['functions'])
        else:
            self._write_data(fileno, '* None')

    def _write_class_function(self, fileno, functions):
        for function in functions:
            if function.startswith('_'):
                continue

            # class method.
            data = '('
            for arg in functions[function]['args']:
                if data != '(':
                    data += ', '
                data += arg
            data += ')'
            self._write_data(fileno, '## {}{}'.format(function, data))
            # method description.
            self._write_data(fileno, '### Function Description')
            method_doc = functions[function]['doc']
            if not method_doc:
                docs = None
                args = None
                rets = None
                examples = None
            else:
                docs, args, rets, examples = self._get_function_desc(
                    functions[function]['doc'])
            if docs:
                for doc in docs:
                    self._write_data(fileno, doc)
            else:
                self._write_data(fileno, '* None')
            # method argument.
            self._write_data(fileno, '### Parameters')
            if args:
                for arg in args:
                    self._write_data(fileno, '* {}'.format(arg))
            else:
                self._write_data(fileno, '* None')
            # method returns.
            self._write_data(fileno, '### Returns')
            if rets:
                for ret in rets:
                    self._write_data(fileno, ret)
            else:
                self._write_data(fileno, '* None')
            # method example.
            self._write_data(fileno, '### Examples')
            if examples:
                for example in examples:
                    self._write_data(fileno, example)
            else:
                self._write_data(fileno, '* None')
            self._write_data(fileno, '')

    def _writer(self, path, name, data):
        assert isinstance(path, str)
        assert isinstance(name, str)
        assert isinstance(data, dict)

        f_markdown = open('{}{}{}.md'.format(
            path, os.sep, name), 'w+', encoding=self.encoding)
        # Model
        self._write_module(f_markdown, data)
        # Desription
        self._write_module_desc(f_markdown, data)
        # Vendor
        self._write_vendor(f_markdown, data)
        # Class Name
        self._write_class(f_markdown, data)

    def _writer_to_markdown(self, docs, files):
        for file in docs:
            if file not in files:
                continue
            _path = os.path.split(file)[0]
            _name = os.path.split(file)[1].split('.')[0]
            if _name.startswith('_'):
                continue
            self._writer(_path, _name, docs[file])

    def transfer(self) -> None:
        '''
        Transfer docstring(google style) to markdown(Archon docs format).

        Examples:
            fullpath = "D:\\project\\test\\docs\\test3"
            docs2md = Docs2Markdown(fullpath)
            docs2md.transfer()
        '''

        if not os.path.isdir(self.fullpath):
            raise TypeError(
                "{} neither dir nor whl format!".format(self.fullpath))

        files = self._get_files(self.fullpath)
        files_to_md = self._check_md_files(files)
        if not files_to_md:
            return
        docs = {}
        for file in files:
            docs[file] = self._walk_python(file)

        # 查找本地依赖关系，把基类的接口中的docstring也转换成markdown格式。
        class_file = self._class_file(docs)
        self._find_bases(docs, class_file)
        docs = self._merge_sub_doc(docs, class_file)
        self._writer_to_markdown(docs, files_to_md)


class Docs2Json(BaseDocsTranser):
    __args_start_re = regexpCompile(r"^(\s*(?:(?:Keyword\s+)?"
                                    r"(?:A|Kwa)rg(?:ument)?|Attribute)s?"
                                    r"\s*:\s*)$", IGNORECASE)
    __args_re = regexpCompile(r"^\s*(?P<name>\w+)\s*(?P<type>\(?\S*\)?)?\s*"
                              r"(?:-|:)+\s+(?P<desc>.+)$")
    __returns_start_re = regexpCompile(
        r"^\s*(?:Return|Yield)s?:\s*$", IGNORECASE)
    __raises_start_re = regexpCompile(r"^\s*(Raises|Exceptions|See Also):\s*$",
                                      IGNORECASE)
    __examples_start_re = regexpCompile(r"^\s*(?:Example|Doctest)s?:\s*$",
                                        IGNORECASE)
    __class_start_en_re = regexpCompile(r"^\s*(##)\s*(C|c(lass))\s+(N|n(ame))\s*$",
                                        IGNORECASE)
    __class_start_cn_re = regexpCompile(r"^\s*(##)\s*(类名)\s*$",
                                        IGNORECASE)
    __function_start_re = regexpCompile(r"^\s*##\s*\w+\((\w+(\,\s*\w+)*)?\)\s*$",
                                        IGNORECASE)
    __function_desc_en_re = regexpCompile(r"^\s*(###)\s*(Function)\s+(Description)\s*$",
                                          IGNORECASE)
    __function_desc_cn_re = regexpCompile(
        r"^\s*(###)\s*(接口描述)\s*$", IGNORECASE)
    __function_parameter_en_re = regexpCompile(
        r"^\s*(###)\s*(Parameters)\s*$", IGNORECASE)
    __function_parameter_cn_re = regexpCompile(
        r"^\s*(###)\s*(参数)\s*$", IGNORECASE)
    __function_return_en_re = regexpCompile(
        r"^\s*(###)\s*(Returns)\s*$", IGNORECASE)
    __function_return_cn_re = regexpCompile(
        r"^\s*(###)\s*(返回值)\s*$", IGNORECASE)
    __function_example_en_re = regexpCompile(
        r"^\s*(###)\s*(Examples)\s*$", IGNORECASE)
    __function_example_cn_re = regexpCompile(
        r"^\s*(###)\s*(例子)\s*$", IGNORECASE)
    __value_none_re = regexpCompile(r"^\s*(\*)\s*(None)\s*$", IGNORECASE)

    def __init__(self, fullpath: str, module: str = None, docspath: str = None) -> None:
        super().__init__(fullpath, module)
        self.docspath = docspath if docspath else "{0}{1}config".format(
            Runtime.get_project_path(), os.sep)

    @classmethod
    def get_function_desc(cls, data: dict):
        desc, arg, returns, example = cls._get_function_desc(cls, data)
        args = []
        for line in arg.splitlines():
            if Docs2Json.__args_re.match(line):
                args.append(line)
            else:
                if args and args[-1]:
                    args[-1] += '\n'
                    args[-1] += line
                else:
                    args.append("")
                    args[-1] += line

        return desc, args, returns, example

    def _get_function_desc(self, data: dict):
        desc = ""
        args = ""
        ret = ""
        example = ""

        step = TransferDef.DOCS_STEP1
        if not data['doc']:
            return desc, args, ret, example

        for line in data['doc'].splitlines():
            line = line.strip()
            if not line:
                continue

            match = Docs2Json.__args_start_re.match(line)
            if match:
                if step != TransferDef.DOCS_STEP2:
                    step = TransferDef.DOCS_STEP2
                    continue
            '''
            match = Docs2Markdown.__args_re.match(line)
            if match:
                args.append(line)
                if step != 'args':
                    step = 'args'
                continue
            '''

            match = Docs2Json.__returns_start_re.match(line)
            if match:
                if step != TransferDef.DOCS_STEP3:
                    step = TransferDef.DOCS_STEP3
                    continue

            match = Docs2Json.__raises_start_re.match(line)
            if match:
                if step != TransferDef.DOCS_STEP4:
                    step = TransferDef.DOCS_STEP4
                    continue

            match = Docs2Json.__examples_start_re.match(line)
            if match:
                if step != TransferDef.DOCS_STEP5:
                    step = TransferDef.DOCS_STEP5
                    continue

            if step == TransferDef.DOCS_STEP1:
                if desc:
                    desc += '\n'
                desc += line
            elif step == TransferDef.DOCS_STEP2:
                if args:
                    args += '\n'
                args += line
            elif step == TransferDef.DOCS_STEP3:
                if ret:
                    ret += '\n'
                ret += line
            elif step == TransferDef.DOCS_STEP5:
                if example:
                    example += '\n'
                example += line

        return desc, args, ret, example

    def _get_arg_desc(self, arg, args, docs):
        if not docs:
            return None
        arg_re = regexpCompile(r"^\s*(?:{})s?:\s*.*$".format(arg), IGNORECASE)
        args_re = []
        for sub in args:
            if arg == sub:
                continue
            args_re.append(regexpCompile(
                r"^\s*(?:{})s?:\s*.*$".format(sub), IGNORECASE))
        desc = ""
        for line in docs.splitlines():
            if desc:
                for re in args_re:
                    if re.match(line):
                        return desc if desc else None
                desc += '\n'
                desc += line
            else:
                if arg_re.match(line):
                    desc = line
        return desc if desc else None

    def filter_str(self, sentence):
        sentence = re.sub(TransferDef.REMOVE_NOTA, '', sentence)
        sentence = sentence.translate(TransferDef.REMOVE_PUNCTUATION_MAP)
        return sentence.strip()

    def _language(self, data):
        assert isinstance(data, str)

        filter_data = self.filter_str(data)
        filter_data = re.sub('[0-9]', '', filter_data).strip()
        if not re.sub('[a-zA-Z]', '', filter_data).strip():
            return TransferDef.LANGUAGE_EN

        for cdata in data:
            if '\u4e00' <= cdata <= '\u9fff':
                return TransferDef.LANGUAGE_CN

        return TransferDef.LANGUAGE_UNKOWN

    def _get_function_doc(self, name, doc):
        # 中英文
        method_doc = {TransferDef.LANGUAGE_EN: {}, TransferDef.LANGUAGE_CN: {}}
        desc, args, ret, example = self._get_function_desc(doc)
        method_doc[TransferDef.LANGUAGE_EN]['name'] = name
        method_doc[TransferDef.LANGUAGE_CN]['name'] = name

        if desc:
            language_type = self._language(desc)
            if language_type == TransferDef.LANGUAGE_EN:
                method_doc[TransferDef.LANGUAGE_EN]['desc'] = desc
            elif language_type == TransferDef.LANGUAGE_CN:
                method_doc[TransferDef.LANGUAGE_CN]['desc'] = desc

        for arg in doc['args']:
            arg_doc_en = {}
            arg_doc_cn = {}
            arg_doc_en['name'] = arg
            arg_doc_cn['name'] = arg
            if 'type' in doc['args'][arg]:
                arg_doc_en['type'] = doc['args'][arg]['type']
                arg_doc_cn['type'] = doc['args'][arg]['type']
            if 'default' in doc['args'][arg]:
                arg_doc_en['default'] = doc['args'][arg]['default']
                arg_doc_cn['default'] = doc['args'][arg]['default']
            arg_desc = self._get_arg_desc(arg, doc['args'], args)
            if arg_desc:
                language_type = self._language(arg_desc)
                if language_type == TransferDef.LANGUAGE_EN:
                    arg_doc_en['desc'] = arg_desc
                elif language_type == TransferDef.LANGUAGE_CN:
                    arg_doc_cn['desc'] = arg_desc
            if 'args' not in method_doc[TransferDef.LANGUAGE_EN]:
                method_doc[TransferDef.LANGUAGE_EN]['args'] = []
            if 'args' not in method_doc[TransferDef.LANGUAGE_CN]:
                method_doc[TransferDef.LANGUAGE_CN]['args'] = []

            method_doc[TransferDef.LANGUAGE_EN]['args'].append(arg_doc_en)
            method_doc[TransferDef.LANGUAGE_CN]['args'].append(arg_doc_cn)

        # 返回值
        if 'returns' in doc:
            language_type = self._language(doc['returns'])
            if language_type == TransferDef.LANGUAGE_EN:
                if 'returns' not in method_doc[TransferDef.LANGUAGE_EN]:
                    method_doc[TransferDef.LANGUAGE_EN]['returns'] = {}
                method_doc[TransferDef.LANGUAGE_EN]['returns']['type'] = doc['returns']
            elif language_type == TransferDef.LANGUAGE_CN:
                if 'returns' not in method_doc[TransferDef.LANGUAGE_CN]:
                    method_doc[TransferDef.LANGUAGE_CN]['returns'] = {}
                method_doc[TransferDef.LANGUAGE_CN]['returns']['type'] = doc['returns']
        if ret:
            language_type = self._language(ret)
            if language_type == TransferDef.LANGUAGE_EN:
                if 'returns' not in method_doc[TransferDef.LANGUAGE_EN]:
                    method_doc[TransferDef.LANGUAGE_EN]['returns'] = {}
                method_doc[TransferDef.LANGUAGE_EN]['returns']['desc'] = ret
            elif language_type == TransferDef.LANGUAGE_CN:
                if 'returns' not in method_doc[TransferDef.LANGUAGE_CN]:
                    method_doc[TransferDef.LANGUAGE_CN]['returns'] = {}
                method_doc[TransferDef.LANGUAGE_CN]['returns']['desc'] = ret
        # 例子
        if example:
            language_type = self._language(example)
            if language_type == TransferDef.LANGUAGE_EN:
                method_doc[TransferDef.LANGUAGE_EN]['examples'] = example
            elif language_type == TransferDef.LANGUAGE_CN:
                method_doc[TransferDef.LANGUAGE_CN]['examples'] = example

        return method_doc

    def _get_class_doc(self, name, doc):
        class_doc = {TransferDef.LANGUAGE_EN: {}, TransferDef.LANGUAGE_CN: {}}
        class_doc[TransferDef.LANGUAGE_EN]['name'] = name
        class_doc[TransferDef.LANGUAGE_CN]['name'] = name
        if doc['doc']:
            language_type = self._language(doc['doc'])
            if language_type == TransferDef.LANGUAGE_EN:
                class_doc[TransferDef.LANGUAGE_EN]['desc'] = doc['doc']
            elif language_type == TransferDef.LANGUAGE_CN:
                class_doc[TransferDef.LANGUAGE_CN]['desc'] = doc['doc']
        if '__init__' in doc['functions']:
            init_desc = self._get_function_doc(
                '__init__', doc['functions']['__init__'])
            if 'args' in init_desc[TransferDef.LANGUAGE_EN]:
                class_doc[TransferDef.LANGUAGE_EN]['args'] = init_desc[TransferDef.LANGUAGE_EN]['args']
            if 'args' in init_desc[TransferDef.LANGUAGE_CN]:
                class_doc[TransferDef.LANGUAGE_CN]['args'] = init_desc[TransferDef.LANGUAGE_CN]['args']

        for function in doc['functions']:
            if function == '__init__':
                continue
            if 'methods' not in class_doc[TransferDef.LANGUAGE_EN]:
                class_doc[TransferDef.LANGUAGE_EN]['methods'] = []
            if 'methods' not in class_doc[TransferDef.LANGUAGE_CN]:
                class_doc[TransferDef.LANGUAGE_CN]['methods'] = []
            function_doc = self._get_function_doc(
                function, doc['functions'][function])
            class_doc[TransferDef.LANGUAGE_EN]['methods'].append(
                function_doc[TransferDef.LANGUAGE_EN])
            class_doc[TransferDef.LANGUAGE_CN]['methods'].append(
                function_doc[TransferDef.LANGUAGE_CN])
        return class_doc

    def _docs_transfer_to_json(self, docs):
        '''
        转换成docs.json的格式
        '''
        #module_docs_en['name'] = self.module
        #module_docs_cn['name'] = self.module
        for file in docs:
            # 以 _ 开头的文件为内部文件，不发布对外接口
            if os.path.split(file)[1].startswith('_'):
                continue
            for sub in docs[file]:
                if docs[file][sub]['type'] == 'class':
                    if 'classes' not in self.vaild_en_docs:
                        self.vaild_en_docs['classes'] = []
                    if 'classes' not in self.vaild_cn_docs:
                        self.vaild_cn_docs['classes'] = []
                    class_docs = self._get_class_doc(sub, docs[file][sub])
                    self.vaild_en_docs['classes'].append(
                        class_docs[TransferDef.LANGUAGE_EN])
                    self.vaild_cn_docs['classes'].append(
                        class_docs[TransferDef.LANGUAGE_CN])
                else:
                    if 'methods' not in self.vaild_en_docs:
                        self.vaild_en_docs['methods'] = []
                    if 'methods' not in self.vaild_cn_docs:
                        self.vaild_cn_docs['methods'] = []
                    function_docs = self._get_function_doc(
                        sub, docs[file][sub])
                    self.vaild_en_docs['methods'].append(
                        function_docs[TransferDef.LANGUAGE_EN])
                    self.vaild_cn_docs['methods'].append(
                        function_docs[TransferDef.LANGUAGE_CN])

    def _docstring_to_json(self):
        docs = {}
        for file in self.python_files:
            docs[file] = self._walk_python(file)

        # 查找本地依赖关系，把基类的接口中的docstring也转换成markdown格式。
        files = self._class_file(docs)
        self._find_bases(docs, files)
        self._merge_sub_doc(docs, files)
        self._docs_transfer_to_json(docs)

    def _writer_to_docfile(self):
        # 修改英文docs文件
        if not os.path.exists(self.docspath):
            os.makedirs(self.docspath)
        if not os.path.isdir(self.docspath):
            raise FileExistsError("self.docspath does not path!")
        en_file = '{}{}{}'.format(
            self.docspath, os.sep, TransferDef.EN_JSON_FILE)
        if os.path.isfile(en_file):
            with open(en_file, "r", encoding='utf-8') as file:
                self.all_en_docs = json.load(file)
        else:
            self.all_en_docs = []
        for doc in self.all_en_docs:
            if self.module == doc['name']:
                self.all_en_docs.remove(doc)
        self.vaild_en_docs['name'] = self.module
        self.all_en_docs.append(self.vaild_en_docs)

        # 修改中文docs文件
        zh_file = '{}{}{}'.format(
            self.docspath, os.sep, TransferDef.CN_JSON_FILE)
        if os.path.isfile(zh_file):
            with open(zh_file, "r", encoding='utf-8') as file:
                self.all_cn_docs = json.load(file)
        else:
            self.all_cn_docs = []
        for doc in self.all_cn_docs:
            if self.module == doc['name']:
                self.all_cn_docs.remove(doc)
        self.vaild_cn_docs['name'] = self.module
        self.all_cn_docs.append(self.vaild_cn_docs)

        # 中文
        with open(en_file, 'w+', encoding='utf-8') as file:
            json.dump(self.all_en_docs, file, indent=4, ensure_ascii=False)

        # 英文
        with open(zh_file, 'w+', encoding='utf-8') as file:
            json.dump(self.all_cn_docs, file, indent=4, ensure_ascii=False)

    def _method_markdown_transfer_to_json(self, markdown: list, language) -> dict:
        assert isinstance(markdown, list)
        if language == TransferDef.LANGUAGE_EN:
            __function_start_re = Docs2Json.__function_start_re
            __function_desc_re = Docs2Json.__function_desc_en_re
            __function_parameter_re = Docs2Json.__function_parameter_en_re
            __function_return_re = Docs2Json.__function_return_en_re
            __function_example_re = Docs2Json.__function_example_en_re
        else:
            __function_start_re = Docs2Json.__function_start_re
            __function_desc_re = Docs2Json.__function_desc_cn_re
            __function_parameter_re = Docs2Json.__function_parameter_cn_re
            __function_return_re = Docs2Json.__function_return_cn_re
            __function_example_re = Docs2Json.__function_example_cn_re

        step = None
        method = {}
        for line in markdown:
            line = line.strip()
            if not line:
                continue
            if __function_start_re.match(line):
                step = TransferDef.MD_SUB_STEP1
            elif __function_desc_re.match(line):
                if step != TransferDef.MD_SUB_STEP2:
                    step = TransferDef.MD_SUB_STEP2
                continue
            elif __function_parameter_re.match(line):
                if step != TransferDef.MD_SUB_STEP3:
                    step = TransferDef.MD_SUB_STEP3
                continue
            elif __function_return_re.match(line):
                if step != TransferDef.MD_SUB_STEP4:
                    step = TransferDef.MD_SUB_STEP4
                continue
            elif __function_example_re.match(line):
                if step != TransferDef.MD_SUB_STEP5:
                    step = TransferDef.MD_SUB_STEP5
                continue

            if step == TransferDef.MD_SUB_STEP1:
                match = re.match(r"^\s*##\s*\w+(\()?", line)
                if match:
                    span = match.span()
                    method['name'] = line[span[0]: span[1]].strip(
                        '##').strip('(').strip()
            elif step == TransferDef.MD_SUB_STEP2:
                if 'desc' not in method:
                    method['desc'] = ""
                if method['desc']:
                    method['desc'] += '\n'
                method['desc'] += line.lstrip()
            elif step == TransferDef.MD_SUB_STEP3:
                if Docs2Json.__value_none_re.match(line):
                    continue
                match = re.match(r'^\s*\*\s*\w+\s*(\(\w*\))?\s*(：|:)', line)
                if match:
                    span = match.span()
                    if 'args' not in method:
                        method['args'] = []
                    args_doc = {}
                    args_doc['name'] = line[span[0]: span[1]].strip(
                        '*').strip('(:|：)').strip()
                    args_doc['desc'] = line[span[1]:].strip()
                    method['args'].append(args_doc)
                else:
                    if 'args' in method:
                        if method['args'][-1]['desc']:
                            method['args'][-1]['desc'] += '\n'
                        method['args'][-1]['desc'] += line.lstrip()
            elif step == TransferDef.MD_SUB_STEP4:
                if Docs2Json.__value_none_re.match(line):
                    continue
                if 'returns' not in method:
                    method['returns'] = {}
                if 'desc' not in method['returns']:
                    method['returns']['desc'] = ""
                if method['returns']['desc']:
                    method['returns']['desc'] += '\n'
                method['returns']['desc'] += line.lstrip().lstrip('*').lstrip()
            elif step == TransferDef.MD_SUB_STEP5:
                if Docs2Json.__value_none_re.match(line):
                    continue
                if 'examples' not in method:
                    method['examples'] = ''
                if method['examples']:
                    method['examples'] += '\n'
                method['examples'] += line.lstrip()
        if 'examples' in method:
            method['examples'] = method['examples'].strip('`').strip()
        return method

    def _class_markdown_transfer_to_json(self, markdown: list, language) -> dict:
        assert isinstance(markdown, list)
        if language == TransferDef.LANGUAGE_EN:
            __class_start_re = Docs2Json.__class_start_en_re
            __function_desc_re = Docs2Json.__function_desc_en_re
            __function_parameter_re = Docs2Json.__function_parameter_en_re
            __function_return_re = Docs2Json.__function_return_en_re
            __function_example_re = Docs2Json.__function_example_en_re
        else:
            __class_start_re = Docs2Json.__class_start_cn_re
            __function_desc_re = Docs2Json.__function_desc_cn_re
            __function_parameter_re = Docs2Json.__function_parameter_cn_re
            __function_return_re = Docs2Json.__function_return_cn_re
            __function_example_re = Docs2Json.__function_example_cn_re

        # 解析类
        classes = []
        step = None
        for class_docs in markdown:
            classes.append({})
            # 解析类
            for line in class_docs['class']:
                if __class_start_re.match(line):
                    step = TransferDef.MD_SUB_STEP1
                    continue
                elif __function_desc_re.match(line):
                    if step != TransferDef.MD_SUB_STEP2:
                        step = TransferDef.MD_SUB_STEP2
                    continue
                elif __function_parameter_re.match(line):
                    if step != TransferDef.MD_SUB_STEP3:
                        step = TransferDef.MD_SUB_STEP3
                    continue
                elif __function_return_re.match(line):
                    if step != TransferDef.MD_SUB_STEP4:
                        step = TransferDef.MD_SUB_STEP4
                    continue
                elif __function_example_re.match(line):
                    if step != TransferDef.MD_SUB_STEP5:
                        step = TransferDef.MD_SUB_STEP5
                    continue

                if step == TransferDef.MD_SUB_STEP1:
                    if re.match(r'^\s*\*\s+\w+', line):
                        classes[-1]['name'] = line.strip('*').strip()
                elif step == TransferDef.MD_SUB_STEP3:
                    if Docs2Json.__value_none_re.match(line):
                        continue
                    match = re.match(
                        r'^\s*\*\s*\w+\s*(\(\w*\))?\s*(：|:)', line)
                    if match:
                        span = match.span()
                        if 'args' not in classes[-1]:
                            classes[-1]['args'] = []
                        args_doc = {}
                        args_doc['name'] = line[span[0]: span[1]].strip(
                            '*').strip('(：|:)').strip()
                        args_doc['desc'] = line[span[1]:].strip()
                        classes[-1]['args'].append(args_doc)
                    else:
                        if 'args' in classes[-1]:
                            if 'desc' in classes[-1]['args'][-1]:
                                classes[-1]['args'][-1]['desc'] += '\n'
                            else:
                                classes[-1]['args'][-1]['desc'] = ''
                            classes[-1]['args'][-1]['desc'] += line.lstrip()
                elif step == TransferDef.MD_SUB_STEP5:
                    if Docs2Json.__value_none_re.match(line):
                        continue
                    if 'examples' not in classes[-1]:
                        classes[-1]['examples'] = ''
                    if classes[-1]['examples']:
                        classes[-1]['examples'] += '\n'
                    classes[-1]['examples'] += line.lstrip()
            if 'examples' in classes[-1]:
                classes[-1]['examples'] = classes[-1]['examples'].strip('`')

            # 解析类方法
            if 'methods' in class_docs:
                classes[-1]['methods'] = []
                for method_docs in class_docs['methods']:
                    method = self._method_markdown_transfer_to_json(
                        method_docs, language)
                    classes[-1]['methods'].append(method)

        return classes

    def _markdown_transfer_to_json(self, markdown, language):
        if not markdown:
            return

        if language == TransferDef.LANGUAGE_EN:
            __class_start_re = Docs2Json.__class_start_en_re
        else:
            __class_start_re = Docs2Json.__class_start_cn_re

        step = None
        """
        [
            [method string], [...]
        ]
        """
        methods_md = []
        '''
        [
            {
                "class": [class string],
                "methods": [[method string]],
            }
        ]
        '''
        classes_md = []

        # 1. 区分类和独立方法
        for line in markdown.splitlines():
            if __class_start_re.match(line):
                if step != TransferDef.MD_STEP2:
                    step = TransferDef.MD_STEP2
                classes_md.append({})

            elif Docs2Json.__function_start_re.match(line):
                is_skip = False
                for skip in TransferDef.MARKDOWN_CN_SKIP:
                    __second_level_skip_re = regexpCompile(
                        r"^\s*(##)\s+({})\s*$".format(skip), IGNORECASE)
                    if __second_level_skip_re.match(line):
                        is_skip = True
                        break
                if not is_skip:
                    if step == TransferDef.MD_STEP2 or step == TransferDef.MD_STEP3:
                        if step == TransferDef.MD_STEP2:
                            step = TransferDef.MD_STEP3
                        if 'methods' not in classes_md[-1]:
                            classes_md[-1]['methods'] = []
                        classes_md[-1]['methods'].append([])
                    # function
                    else:
                        if step != TransferDef.MD_STEP1:
                            step = TransferDef.MD_STEP1
                        methods_md.append([])

            if step == TransferDef.MD_STEP1:
                methods_md[-1].append(line)
            elif step == TransferDef.MD_STEP2:
                if 'class' not in classes_md[-1]:
                    classes_md[-1]['class'] = []
                classes_md[-1]['class'].append(line)
            elif step == TransferDef.MD_STEP3:
                classes_md[-1]['methods'][-1].append(line)

        # 2. 解析类的内容
        classes_docs = None
        if classes_md:
            classes_docs = self._class_markdown_transfer_to_json(
                classes_md, language)

        # 3. 解析方法的内容
        methods_docs = []
        if methods_md:
            for method_md in methods_md:
                method = self._method_markdown_transfer_to_json(
                    method_md, language)
                methods_docs.append(method)

        if language == TransferDef.LANGUAGE_EN:
            if classes_docs:
                if 'classes' in self.vaild_en_docs:
                    self.vaild_en_docs['classes'] += classes_docs
                else:
                    self.vaild_en_docs['classes'] = classes_docs
            if methods_docs:
                if 'methods' in self.vaild_en_docs:
                    self.vaild_en_docs['methods'] += methods_docs
                else:
                    self.vaild_en_docs['methods'] = methods_docs
        else:
            if classes_docs:
                if 'classes' in self.vaild_cn_docs:
                    self.vaild_cn_docs['classes'] += classes_docs
                else:
                    self.vaild_cn_docs['classes'] = classes_docs
            if methods_docs:
                if 'methods' in self.vaild_cn_docs:
                    self.vaild_cn_docs['methods'] += methods_docs
                else:
                    self.vaild_cn_docs['methods'] = methods_docs

    def _markdown_to_json(self):
        for md_file in self.md_files:
            data = self._read_file(md_file)
            language = self._language(data)
            if language == TransferDef.LANGUAGE_EN:
                self._markdown_transfer_to_json(data, TransferDef.LANGUAGE_EN)
            elif language == TransferDef.LANGUAGE_CN:
                self._markdown_transfer_to_json(data, TransferDef.LANGUAGE_CN)

    def transfer(self) -> None:
        '''
        Transfer docstring(google style) to markdown(Archon docs format).

        Examples:
            fullpath = "D:\\project\\archon\\plugins\\archon\\archon"
            doc2js = Docs2Json(fullpath)
            doc2js.transfer()
        '''

        if not os.path.isdir(self.fullpath):
            raise FileNotFoundError(
                "{} not found".format(self.fullpath))

        # 查找模块的文件
        self._get_files(self.fullpath)

        # python中的docstring转成json
        self._docstring_to_json()

        # markdown 转成json
        self._markdown_to_json()

        # 写入文档
        self._writer_to_docfile()
