import re
import socket
import select
import time
from typing import Tuple
from ._tcp import _TCPMessage


class TCPDef:
    RECV_LEN = 65535


class TCPClient(_TCPMessage):
    """TCPClinet is support for setup a tcp client connet to server.

    Args:
        partno (str): Device PartNo.
        CRLF (str): TCP Message endle, also you can reserve blank and
                    send command with endle every time.
        setting (str): TCP clinet connect setting. Format is ip:port.

    Attributes:
        ipaddr (str): TCP connect ip.
        port (int): TCP connect port.

    Examples:
        >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = "") -> None:
        self.partno = partno
        self.crlf = CRLF
        self._opened = False
        if not re.match(r"(^(((2[0-4]\d|25[0-5]|[1-9]?\d|1\d{2})\.){3}(2[0-4]\d|25[0-5]|[01]?\d\d?):"
                        r"(6553[0-5]|655[0-2][0-9]|65[0-4]\d{2}|6[0-4]\d{3}|[1-5]\d{4}|[1-9]\d{0,3}))$)",
                        setting.strip()):
            raise ValueError('Setting invalid: {}'.format(setting))

        _parameters = setting.strip().split(':')
        self.ipaddr = _parameters[0].strip()
        self.port = int(_parameters[1].strip())
        self._timeout = None

    def open(self):
        """
        """
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._sock.connect((self.ipaddr, self.port))
        self._sock.setblocking(False)
        self._opened = True

    def close(self):
        """
        """
        try:
            self._sock.shutdown(1)
            self._sock.close()
        except Exception:
            pass
        self._sock = None
        self._opened = False

    def reopen(self):
        try:
            self._sock.shutdown(1)
            self._sock.close()
        except Exception:
            pass
        self.open()

    @property
    def is_open(self):
        return self._opened

    @property
    def timeout(self) -> Tuple[int, float]:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Tuple[int, float]) -> None:
        if timeout:
            assert isinstance(timeout, (int, float))
        self._timeout = timeout

    def write(self, command: str) -> None:
        """Send command to TCP server.

        Args:
            command (str): command to TCP server.

        Examples:
            >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
            tcp.send('[]version(255)')
        """
        assert isinstance(command, str)

        return self.send(command)

    def send(self, command: Tuple[str, bytes]) -> None:
        """Send command to TCP server.

        Args:
            command (str): command to TCP server.

        Examples:
            >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
            tcp.write('[]version(255)')
        """
        assert isinstance(command, str)
        try:
            ret = self._sock.send(
                ('{}{}'.format(command, self.crlf)).encode('utf-8'))
        except ConnectionResetError as _except:
            self.close()
            raise _except
        return ret

    def read(self) -> str:
        """Recive reply from TCP server.

        Returns:
            str: Reply data from server.

        Examples:
            >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
            tcp.send('[]version(255)')
            print(tcp.read())
        """
        return self.recv()

    def read_bytes(self) -> bytes:
        """Recive reply from TCP server.

        Returns:
            bytes: Reply data from server.

        Examples:
            >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
            tcp.send('[]version(255)')
            print(tcp.read_bytes())
        """
        rsocket = [self._sock]
        wsocket = []
        xsocket = []
        data = b''
        start_time = time.time()
        while self.is_open:
            rlist, _, _ = select.select(rsocket, wsocket, xsocket, 0.1)
            for sock in rlist:
                if sock != self._sock:
                    continue
                while True:
                    try:
                        recv_data = sock.recv(TCPDef.RECV_LEN)
                    except BlockingIOError:
                        break

                    if recv_data:
                        data += recv_data

                    if self._timeout and self._timeout < time.time() - start_time:
                        raise TimeoutError(
                            'Recive reply from TCP server timeout!')

                if not data:
                    self.close()
                    return data

            if self._timeout and self._timeout < time.time() - start_time:
                raise TimeoutError('Recive reply from TCP server timeout!')

            if not rlist:
                continue

            return data

    def recv(self) -> str:
        """Recive reply from TCP server.

        Returns:
            str: Reply data from server.

        Examples:
            >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
            tcp.send('[]version(255)')
            print(tcp.recv())
        """
        return self.read_bytes().decode('utf-8')

    def query(self, command: str) -> str:
        """Send command to TCP server, then recive reply data.

        Args:
            command (str): command to TCP server.

        Returns:
            str: Reply data from server.

        Examples:
            >>> tcp = TCPClient(setting='169.254.1.32:7600', CRLF='\\r\\n')
            print(tcp.query('[]version(255)'))
        """
        assert isinstance(command, str)

        self.send(command)
        return self.recv()
