import os
import re
from typing import Any
from ._rpc import _RPC
from tinyddp.client import DDPClient
from ._rpc import CommandException
from ._rpc import RPCBase


class TinyDDPDef:
    XAVIER_CONTROL_BIT = '/app/project/config/controlbit.json'
    XAVIER_PROFILE = '/app/project/config/profile.json'
    DEFAULT_TIMEOUT = 0xFFFFFFF
    USER = 'root'
    PASSWD = '123456'
    DOCS_ZH_NAME = 'docstring_zh.json'
    DOCS_EN_NAME = 'docstring_en.json'
    SYS_DOCS_ZH_FILE = os.path.abspath(
        "{0}{1}..{1}config{1}{2}".format(os.getcwd(), os.sep, DOCS_ZH_NAME))
    SYS_DOCS_EN_FILE = os.path.abspath(
        "{0}{1}..{1}config{1}{2}".format(os.getcwd(), os.sep, DOCS_EN_NAME))
    VERSION = '/app/version.json'

    MGNT_SERVER = 'mgnt'
    MGNT_PORT = 7800
    CHECKSUM_FILE = 'checksum.json'


class _TinyDDPClient(_RPC):
    """TinyDDP is support for setup ddp connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Reserve.
        setting (str): DDP connect setting. Format is "ip:port".

    Attributes:
        ipaddr (str): DDP connect ip.
        port (int): DDP connect port.

    Examples:
        >>> ddp = TinyDDP(partno="Xavier-001-007A", setting="169.254.1.32:7801')
    """

    def __init__(self, ipaddr: str, port: int) -> None:
        _RPC.__init__(self, ipaddr, port)
        self._ssh = None
        self._sftp = None
        self._tran = None

    def open(self) -> None:
        """
        """
        if not self._link:
            self._link = DDPClient('ws://{}:{}'.format(self.ipaddr, self.port))
        # 顺序不能颠倒，否则会导致实例被替换，原因未知
        setattr(self, 'call', self._link.call)
        self._link.connect()
        self._opened = True

    def close(self) -> None:
        """
        """
        if self._link:
            self._link.terminate()
        if hasattr(self, 'call'):
            delattr(self, 'call')
        self._link = None
        self._opened = False

    def reopen(self) -> None:
        if self._link:
            self._link.terminate()
            self._link
        self.open()


class TinyDDP(RPCBase):
    """TinyDDP is support for setup rpc connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Reserve.
        setting (str): DDP connect setting. Format is "ip:port".

    Attributes:
        ipaddr (str): DDP connect ip.
        port (int): DDP connect port.

    Examples:
        >>> rpc = TinyRPC(partno="Xavier-001-007A", setting="169.254.1.32:7801')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = "") -> None:
        RPCBase.__init__(self, partno, setting, CRLF)
        self._xavier_controlbit = TinyDDPDef.XAVIER_CONTROL_BIT
        self._xavier_profile = TinyDDPDef.XAVIER_PROFILE
        self._user = TinyDDPDef.USER
        self._passwd = TinyDDPDef.PASSWD
        self._docs_zh_name = TinyDDPDef.DOCS_ZH_NAME
        self._docs_en_name = TinyDDPDef.DOCS_EN_NAME
        self._sys_docs_en_file = TinyDDPDef.SYS_DOCS_EN_FILE
        self._sys_docs_zh_file = TinyDDPDef.SYS_DOCS_ZH_FILE
        self._version = TinyDDPDef.VERSION
        self._mgnt_server = TinyDDPDef.MGNT_SERVER
        self._mgnt_port = TinyDDPDef.MGNT_PORT
        self._checksum_file = TinyDDPDef.CHECKSUM_FILE
        self._rpc_proxy = TinyDDPProxy
        self._rpc_client = _TinyDDPClient

    def call(self, command: str) -> Any:
        assert isinstance(command, str)

        group = re.match(r'.*\(.*\)$', command.strip())
        if not group:
            raise CommandException(
                'Invalid interface: {}'.format(command))
        group = re.split(r'\(|\)', command)
        return self.link.call(group[0],
                              self._decode_command(group[1]),
                              {},
                              self._timeout if self._timeout else None)

    def _get_version(self, link):
        return link.call(
            "system.fw_version", [], {}, 0.5)

    def _get_all_methods(self, link):
        return link.call(
            "server.all_methods", [], {}, 1)


class TinyDDPProxy():
    def __init__(self, link, prefix) -> None:
        self._link = link
        self._prefix = prefix

    def __getattr__(self, __name: str) -> Any:
        func = lambda *args, **kwargs: self._link.call("{}.{}".format(
            self._prefix, __name), args, kwargs, self._link._timeout if self._link._timeout else None)
        func.__name__ = "{}.{}".format(self._prefix, __name)
        return func
