import os
import re
import time
from typing import Tuple
import serial
from modem import XMODEM
from modem import YMODEM
from modem import ZMODEM


class SerialDef:
    parity = {
        'N': 'none',
        'E': 'even',
        'O': 'odd',
        'M': 'mark',
        'S': 'space'
    }


class Serial():
    """Serial is support for setup uart serial connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Serial Message endle, also you can reserve blank and
                    send command with endle every time.
        setting (str): Serial connect setting. Format is "port, baudrate, bytesize, parity, stopbits".

    Attributes:
        partno (str): instructment partno.
        port (str): Device name.
        baudrate (int): Baud rate such as 9600 or 115200 etc.
        bytesize (int): Number of data bits. Possible values: 5, 6, 7, 8.
        parity (str): Enable parity checking. Possible values: 'N', 'E', 'O', 'M', 'S'
        stopbits (int, float): Number of stop bits. Possible values: 1, 1.5, 2
        ddpclient (object): DDPClient connect to xavier.

    Examples:
        >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = "") -> None:
        self._link = None
        self.partno = partno
        self._opened = False
        self.crlf = CRLF
        self.ddpclient = None
        if re.match(r'^(\d{2,})', setting) or not setting.strip():
            offset = -1
        else:
            offset = 0

        self.port = None
        self.baudrate = 115200
        self.bytesize = 8
        self.parity = 'N'
        self.stopbits = 1
        self._config = False
        self._timeout = None

        if not setting.strip():
            return

        parameters = setting.strip().split(',')
        _param_len = len(parameters)
        if _param_len == 5 + offset:
            if '.' in parameters[4 + offset]:
                self.stopbits = float(parameters[4 + offset])
            else:
                self.stopbits = int(parameters[4 + offset])

        if _param_len >= 4 + offset:
            self.parity = parameters[3 + offset].strip().strip('\'').strip('"')
            if self.parity in SerialDef.parity.values():
                for parity in SerialDef.parity:
                    if SerialDef.parity[parity] == self.parity:
                        self.parity = parity
            if self.parity not in SerialDef.parity:
                raise ValueError('Setting invalid: {}'.format(setting))
        if _param_len >= 3 + offset:
            self.bytesize = int(parameters[2 + offset])
        if _param_len >= 2 + offset:
            self.baudrate = int(parameters[1 + offset])
            self._config = True
        if _param_len >= 1 + offset and offset >= 0:
            self.port = parameters[offset]
        # 每秒接收的数据量 = 波特率/(位宽+停止位)
        self._bytes_per_sec = self.baudrate / (self.bytesize + self.stopbits)
        # 接收一个数据的数据（秒） = （1秒/每秒接收的数据量）
        self._recv_bytes_sec = 1 / self._bytes_per_sec

    def open(self):
        """
        """
        if self.ddpclient:
            self._link = self.ddpclient.get_proxy(self.port)
            if self._config:
                self._link.config(self.baudrate, self.bytesize,
                                  SerialDef.parity[self.parity], self.stopbits)
        else:
            self._link = serial.Serial(
                self.port, self.baudrate, self.bytesize, self.parity,
                self.stopbits, timeout=0, write_timeout=0)
        self._opened = True

    def close(self):
        """
        """
        try:
            self._link.close()
        except Exception:
            pass
        self._link = None
        self._opened = False

    def reopen(self) -> None:
        try:
            self._link.close()
        except Exception:
            pass
        self.open()

    @property
    def device(self):
        return self.port

    @device.setter
    def device(self, port: str):
        assert isinstance(port, str)

        self.port = port

    @property
    def is_open(self):
        return self._opened

    @property
    def remote(self):
        return self.ddpclient

    @remote.setter
    def remote(self, clinet):
        assert clinet.__class__.__name__ in ['TinyDDP', 'TinyRPCX']
        self.ddpclient = clinet

    @property
    def timeout(self) -> Tuple[int, float]:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Tuple[int, float]) -> None:
        if timeout:
            assert isinstance(timeout, (int, float))
        self._timeout = timeout

    def write(self, command: str) -> None:
        """Send message via uart bus.

        Args:
            command (str): command to uart device.

        Examples:
            >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
            uart.send('[]version(255)')

        CSV Examples:
            | device | function |    parameter   |
            | ------ | -------- | -------------- |
            |  uart  |   send   | []version(255) |
        """
        assert isinstance(command, str)
        _cmd = '{}{}'.format(command, self.crlf).encode('utf-8')
        if self.ddpclient:
            _cmd_list = [_ for _ in _cmd]
            return self._link.write_hex(_cmd_list)

        try:
            ret = self._link.write(_cmd)
        except Exception as _except:
            if _except.__class__.__name__ == 'SerialException':
                self.close()
            raise _except
        self._link.flush()
        return ret

    def write_hex(self, command: list) -> None:
        """Send hex message via uart bus.

        Args:
            command (tuple[str, list]): hex command to uart device.

        Examples:
            >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
            uart.write_hex("[0x5B, 0x5D, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6F, 0x6E, 0x28, 0x32, 0x35, 0x35, 0x29, 0x0D, 0x0A]")
        """
        assert isinstance(command, list)

        if self.ddpclient:
            return self._link.write_hex(command)

        _data = b''
        for sub in command:
            _data = _data + (sub).to_bytes(1, 'big')

        for sub in self.crlf:
            _data += sub.encode('utf-8')
        return self.write(_data)

    def read_bytes(self, length: int = None) -> bytes:
        """Recive reply via uart.

        Args:
            length (int): length of recive data.

        Returns:
            bytes: Reply data.

        Examples:
            >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
            uart.write('[]version(255)')
            print(uart.read_bytes())
        """
        datas = b''
        start_time = time.time()
        if self.ddpclient:
            if length:
                data = self._link.read_hex(length)
                for _ in data:
                    datas += _.to_bytes(1, 'big')
            else:
                data_len = 128
                while True:
                    data = self._link.read_hex(data_len)
                    for _ in data:
                        datas += _.to_bytes(1, 'big')
                    if not data and datas:
                        break
                    if self._timeout and self._timeout < time.time() - start_time:
                        raise TimeoutError('Uart communication timeout!')
        else:
            if length:
                data = self._link.read(length)
                for _ in data:
                    datas += _.to_bytes(1, 'big')
            else:
                datas = b''
                idle = 0
                while True:
                    try:
                        data = self._link.read(1)
                    except Exception as _except:
                        if _except.__class__.__name__ == 'SerialException':
                            self.close()
                        raise _except
                    if data:
                        datas += data
                        idle = 0
                    elif datas:
                        if idle >= 40:
                            break
                        idle += 1
                        time.sleep(self._recv_bytes_sec * 1.2)
                    else:
                        time.sleep(self._recv_bytes_sec * 1.2)
                    if self._timeout and self._timeout < time.time() - start_time:
                        raise TimeoutError('Uart communication timeout!')

        return datas

    def read(self, length: int = None) -> str:
        """Recive reply via uart.

        Args:
            length (int): length of recive data.

        Returns:
            str: Reply data.

        Examples:
            >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
            uart.write('[]version(255)')
            print(uart.read())
        """
        if length:
            assert isinstance(length, int)

        return self.read_bytes(length).decode('utf-8')

    def read_hex(self, length: int = None) -> list:
        """Recive reply via uart.

        Args:
            length (int): length of recive data.

        Returns:
            list: Reply data.

        Examples:
            >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
            uart.write_hex("[0x5B, 0x5D, 0x76, 0x65, 0x72, 0x73, 0x69, 0x6F, 0x6E, 0x28, 0x32, 0x35, 0x35, 0x29, 0x0D, 0x0A]")
            print(uart.read_hex())
        """
        if length:
            assert isinstance(length, int)

        return [_ for _ in self.read_bytes(length)]

    def query(self, command: str) -> str:
        """Send command to via uart, then recive reply data.

        Args:
            command (str): command to uart device.

        Returns:
            str: Reply data.

        Examples:
            >>> uart = Serial(partno="IT8812C", setting="COM1,115200,8,N,1", CRLF='\\r\\n')
            print(uart.query('[]version(255)'))
        """
        assert isinstance(command, str)
        self.write(command)
        return self.read()

    def __getc(self, size: int, timeout: int = 5) -> bytes:
        assert isinstance(size, int)
        assert isinstance(timeout, int)

        origin_timeout = self._link.timeout
        self._link.timeout = timeout
        data = self._link.read(size)
        if not isinstance(data, bytes):
            data = data.encode('utf-8')
        self._link.timeout = origin_timeout
        return data

    def __putc(self, data: bytes, timeout: int) -> bool:
        assert isinstance(data, bytes)
        assert isinstance(timeout, int)

        origin_timeout = self._link.write_timeout
        self._link.write_timeout = timeout
        self._link.write(data)
        self._link.write_timeout = origin_timeout

    def sendfile_xmodem(self, file: str) -> bool:
        assert isinstance(file, str)
        assert os.path.isfile(file)

        fileno = open(file, 'rb')
        xmodem = XMODEM(self.__getc, self.__putc)
        xmodem.send(fileno)

    def sendfile_ymodem(self, file: str) -> bool:
        assert isinstance(file, str)
        assert os.path.isfile(file)

        fileno = open(file, 'rb')
        xmodem = YMODEM(self.__getc, self.__putc)
        xmodem.send(fileno)

    def sendfile_zmodem(self, file: str) -> bool:
        assert isinstance(file, str)
        assert os.path.isfile(file)

        fileno = open(file, 'rb')
        xmodem = ZMODEM(self.__getc, self.__putc)
        xmodem.send(fileno)
