import re
import time
from typing import Tuple, Union
import pyvisa
from pyvisa.constants import VI_WRITE_BUF
from pyvisa.constants import Parity


class SerialDef:
    parity = {
        'N': Parity.none,
        'E': Parity.odd,
        'O': Parity.even,
        'M': Parity.mark,
        'S': Parity.space
    }


class Visa():
    """Visa is support for setup visa connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Visa Message endle, also you can reserve blank and
                    send command with endle every time.
        setting (str): Visa connect setting. Format is resource.

    Attributes:
        resource (str): Visa connect resoure.

    Examples:
        >>> visa = Visa(partno="IT8812C", setting='TCPIP0::169.254.1.32::7801::SOCKET', CRLF='\\r\\n')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = None) -> None:
        resource = setting.strip().split(',')[0].strip()
        if re.match(r"(^(ASRL[a-zA-Z0-9\-_]*::.*::.*::INSTR)$)|"
                    r"(^(ASRL[a-zA-Z0-9\-_]*::.*::.*::INSTR),\d{1,}.*)|"
                    r"(^(ASRL[a-zA-Z0-9\-_]*(::INSTR){0,1})$)|"
                    r"(^(ASRL[a-zA-Z0-9\-_]*(::INSTR){0,1}),\d{1,}.*)|"
                    r"^((\d{2,})|(\d{2,}\,[5-8])|"
                    r"(\d{2,}\,[5-8]\,(none)|(even)|(odd)|(mark)|(space))|"
                    r"(\d{2,}\,[5-8]\,(none)|(even)|(odd)|(mark)|(space)\,\d{1})|"
                    r"(\d{2,}\,[5-8]\,(N)|(E)|(O)|(M)|(S))|"
                    r"(\d{2,}\,[5-8]\,(N)|(E)|(O)|(M)|(S)\,\d{1}))", setting.strip()):

            ''' Visa serial resource format
            Interface          |Syntax
            -------------------|----------------------------------------------------------------------------------
            ENET-Serial INSTR  |ASRL[0]::host address::serial port::INSTR
            Serial INSTR       |ASRLboard[::INSTR]
            '''

            self._decode_uart(setting)
        elif self._is_visa_resource(resource):
            self.resource = resource
            self._config = False
        elif not setting.strip():
            self.resource = None
            self._config = False
        else:
            raise ValueError('Setting invalid: {}'.format(setting))

        self.partno = partno
        self.crlf = CRLF
        self._opened = False
        self._timeout = None
        self.read_mode = None

    def _is_visa_resource(self, resource) -> bool:
        '''Visa resource format(except serial)
        Interface          |Syntax
        -------------------|----------------------------------------------------------------------------------
        ENET-Serial INSTR  |ASRL[0]::host address::serial port::INSTR
        Serial INSTR       |ASRLboard[::INSTR]
        GPIB INSTR         |GPIB[board]::primary address[::secondary address][::INSTR]
        GPIB INTFC         |GPIB[board]::INTFC
        PXI BACKPLANE      |PXI[interface]::chassis number::BACKPLANE
        PXI INSTR          |PXI[bus]::device[::function][::INSTR]
        PXI INSTR          |PXI[interface]::bus-device[.function][::INSTR]
        PXI INSTR          |PXI[interface]::CHASSISchassis number::SLOTslot number[::FUNCfunction][::INSTR]
        PXI MEMACC         |PXI[interface]::MEMACC
        Remote NI-VISA     |visa://host address[:server port]/remote resource
        TCPIP INSTR        |TCPIP[board]::host address[::LAN device name][::INSTR]
        TCPIP SOCKET       |TCPIP[board]::host address::port::SOCKET
        USB INSTR          |USB[board]::manufacturer ID::model code::serial number[::USB interface number][::INSTR]
        USB RAW            |USB[board]::manufacturer ID::model code::serial number[::USB interface number]::RAW
        VXI BACKPLANE      |VXI[board][::VXI logical address]::BACKPLANE
        VXI INSTR          |VXI[board]::VXI logical address[::INSTR]
        VXI MEMACC         |VXI[board]::MEMACC
        VXI SERVANT        |VXI[board]::SERVANT
        '''

        if re.match(r"(^(ASRL[a-zA-Z0-9\-_]*::.*::.*::INSTR)$)|"
                    r"(^(ASRL[a-zA-Z0-9\-_]*(::INSTR){0,1})$)|"
                    r"(^(GPIB[a-zA-Z0-9\-_]*::\d{1,}(::\d{1,}){0,1}(::INSTR){0,1})$)|"
                    r"(^(GPIB[a-zA-Z0-9\-_]*::INTFC)$)|"
                    r"(^(PXI[a-zA-Z0-9\-_]*::\d{1,}(::.*){0,1}(::INSTR){0,1})$)|"
                    r"(^(PXI[a-zA-Z0-9\-_]*::[0-9\-_]+(\.[0-9\-_]+){0,1}(::INSTR){0,1})$)|"
                    r"(^(PXI[a-zA-Z0-9\-_]*::\d{1,}::BACKPLANE)$)|"
                    r"(^(PXI[a-zA-Z0-9\-_]*::[a-zA-Z0-9\-_]+::[a-zA-Z0-9\-_]+(::[a-zA-Z0-9\-_]+){0,1}(::INSTR){0,1})$)|"
                    r"(^(PXI[a-zA-Z0-9\-_]*::MEMACC)$)|"
                    r"(^(TCPIP[a-zA-Z0-9\-_]*::[a-zA-Z0-9\-_\.]+(::[a-zA-Z0-9\-_]+){0,1}(::INSTR){0,1})$)|"
                    r"(^(TCPIP[a-zA-Z0-9\-_]*::((2[0-4]\d|25[0-5]|[1-9]?\d|1\d{2})\.){3}(2[0-4]\d|25[0-5]|[01]?\d\d?)::"
                    r"(6553[0-5]|655[0-2][0-9]|65[0-4]\d{2}|6[0-4]\d{3}|[1-5]\d{4}|[1-9]\d{0,3})::SOCKET)$)|"
                    r"(^(visa://[a-zA-Z0-9\-_\.\:]+(:[0-9]+){0,1}/.+)$)|"
                    r"(^(USB[a-zA-Z0-9\-_]*::[a-zA-Z0-9\-_]+::[a-zA-Z0-9\-_]+::[a-zA-Z0-9\-_]+(::[a-zA-Z0-9\-_]+){0,1}(::INSTR){0,1})$)|"
                    r"(^(USB[a-zA-Z0-9\-_]*::[a-zA-Z0-9\-_]+::[a-zA-Z0-9\-_]+::[a-zA-Z0-9\-_]+(::[a-zA-Z0-9\-_]+){0,1}::RAW)$)|"
                    r"(^(VXI[a-zA-Z0-9\-_]*(::[a-zA-Z0-9\-_]+){0,1}::BACKPLANE)$)|"
                    r"(^(VXI[a-zA-Z0-9\-_]*::[a-zA-Z0-9\-_]+(::INSTR){0,1})$)|"
                    r"(^(VXI[a-zA-Z0-9\-_]*::MEMACC)$)|"
                    r"(^(VXI[a-zA-Z0-9\-_]*::SERVANT)$)", resource.strip()):
            return True

        return False

    def _decode_uart(self, setting):
        self.resource = None
        self.baudrate = 115200
        self.bytesize = 8
        self.parity = 'none'
        self.stopbits = 1
        _parameters = setting.strip().split(',')
        _param_len = len(_parameters)
        if re.match(r'^(\d{2,})', setting) or not setting:
            offset = -1
        else:
            offset = 0

        if _param_len == 5 + offset:
            if '.' in _parameters[4 + offset]:
                self.stopbits = float(_parameters[4 + offset])
            else:
                self.stopbits = int(_parameters[4 + offset])

        if _param_len >= 4 + offset:
            parity = _parameters[3 +
                                 offset].strip().strip('\'').strip('"')
            if parity not in SerialDef.parity.keys():
                raise ValueError('Setting invalid: {}'.format(setting))
            self.parity = SerialDef.parity[parity]
        if _param_len >= 3 + offset:
            self.bytesize = int(_parameters[2 + offset])
        if _param_len >= 2 + offset:
            self.baudrate = int(_parameters[1 + offset])
        if _param_len >= 1 + offset and offset >= 0:
            self.resource = _parameters[0]
        self._config = True

    @property
    def device(self):
        return self.resource

    @device.setter
    def device(self, resource):
        assert isinstance(resource, str)

        if not self._is_visa_resource(resource):
            raise ValueError('Resource invalid: {}'.format(resource))

        self.resource = resource.strip()

    @property
    def timeout(self) -> Tuple[int, float]:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Tuple[int, float]) -> None:
        if timeout:
            assert isinstance(timeout, (int, float))
        self._timeout = timeout

    def open(self, visa_library: Union[str, pyvisa.highlevel.VisaLibraryBase] = ''):
        """
        """
        self._resource_manager = pyvisa.ResourceManager(visa_library)
        self._link = self._resource_manager.open_resource(self.resource)
        if self._config:
            self._link.baud_rate = self.baudrate
            self._link.data_bits = self.bytesize
            self._link.parity = self.parity
            self._link.stop_bits = self.stopbits
        if self.crlf:
            self._link.write_termination = self.crlf
        if self._link:
            self._link.timeout = 3000
        self._opened = True

    def close(self):
        """
        """
        try:
            self._link.close()
        except Exception:
            pass
        self._resource_manager = None
        self._link = None
        self._opened = False
        self.read_mode = None

    def reopen(self, visa_library: Union[str, pyvisa.highlevel.VisaLibraryBase] = ''):
        try:
            self._link.close()
        except Exception:
            pass
        self.open(visa_library)

    @property
    def is_open(self):
        return self._opened

    def write(self, command: str) -> None:
        """Send command to via visa.

        Args:
            command (str): command to visa device.

        Examples:
            >>> visa = Visa(partno="IT8812C", setting="TCPIP0::169.254.1.32::7801::SOCKET", CRLF='\\r\\n')
            visa.write('*IDN?')
        """

        assert isinstance(command, str)
        try:
            ret = self._link.write(command)
        except pyvisa.errors.VisaIOError as _except:
            self.close()
            raise _except

        if hasattr(self._link, 'flush'):
            self._link.flush(VI_WRITE_BUF)
        return ret

    def read_bytes(self, length, chunk_size: int = None, break_on_termchar: bool = False) -> bytes:
        """Recive reply via visa.

        Returns:
            bytes: Reply data.

        Examples:
            >>> visa = Visa(partno="IT8812C", setting='TCPIP0::169.254.1.32::7801::SOCKET', CRLF='\\r\\n')
            visa.write('*IDN?')
            print(visa.read_bytes())
        """
        start_time = time.time()
        while True:
            try:
                return self._link.read_bytes(length, chunk_size, break_on_termchar)
            except pyvisa.errors.VisaIOError as _except:
                if 'VI_ERROR_TMO' not in repr(_except):
                    self.close()
                    raise _except
            if self._timeout and self._timeout < time.time() - start_time:
                raise TimeoutError('Read data Time out!')

    def read(self, length: int = None) -> str:
        """Recive reply via visa.

        Returns:
            str: Reply data.

        Examples:
            >>> visa = Visa(partno="IT8812C", setting='TCPIP0::169.254.1.32::7801::SOCKET', CRLF='\\r\\n')
            visa.write('*IDN?')
            print(visa.read())
        """
        start_time = time.time()
        if length:
            return self.read_bytes(length).decode('utf-8')
        datas = b''
        try_cnt = 0
        while True:
            try:
                if self.read_mode != 'all':
                    if self.read_mode == None:
                        self._link.timeout = 100
                    try_cnt += 1
                    data = self._link.read_bytes(1)
                    if data:
                        datas += data
                        if self.read_mode == None:
                            self._link.timeout = 3000
                            self.read_mode = 'bytes'
                        if datas.endswith(b'\n') and datas.strip():
                            return datas.decode('utf-8')
                elif self.read_mode == 'all':
                    return self._link.read()
            except pyvisa.errors.VisaIOError as _except:
                if 'VI_ERROR_TMO' not in repr(_except):
                    self.close()
                    raise _except
            if self._timeout and self._timeout < time.time() - start_time:
                self._link.timeout = 3000
                raise TimeoutError('Read data Time out!')
            if try_cnt > 6 and not self.read_mode:
                self._link.timeout = 3000
                self.read_mode = 'all'

    def query(self, command: str) -> str:
        """Send command to via visa, then recive reply data.

        Args:
            command (str): command to visa device.

        Returns:
            str: Reply data.

        Examples:
            >>> visa = Visa(partno="IT8812C", setting='TCPIP0::169.254.1.32::7801::SOCKET', CRLF='\\r\\n')
            print(visa.query('*IDN?'))
        """
        assert isinstance(command, str)

        self.write(command)
        return self.read()

    def read_raw(self, size: int = None) -> bytes:
        '''
        Read the unmodified string sent from the instrument to the computer.

        In contrast to read(), no termination characters are stripped.

        Args:
            size (int): The chunk size to use to perform the reading.
                        Defaults to None, meaning the resource wide set value is set.

        Returns:
            bytes: read from the instrument.
        '''
        start_time = time.time()
        while True:
            try:
                return self._link.read_raw(size)
            except pyvisa.errors.VisaIOError as _except:
                if 'VI_ERROR_TMO' not in repr(_except):
                    self.close()
                    raise _except
            if self._timeout and self._timeout < time.time() - start_time:
                raise TimeoutError('Read data Time out!')

    def write_raw(self, message: bytes) -> int:
        '''
        Write a byte message to the device.

        Args:
            message (bytes): The message to be sent.
        Returns:
            int: Number of bytes written
        '''
        assert isinstance(message, bytes)

        try:
            self._link.timeout = 30000
            ret = self._link.write_raw(message)
            self._link.timeout = 50
        except pyvisa.errors.VisaIOError as _except:
            self.close()
            raise _except

        if hasattr(self._link, 'flush'):
            self._link.flush(VI_WRITE_BUF)
        return ret
